unit  BackProp;    { unit for Back Propagation Algorithm implementation }
                   { author Petr Bozovsky }
interface

const MaxLayer   =  5;                            { max. number of layers }
      MaxNeuron  = 15;              { max. number of neurons in one layer }
      MaxPattern = 50;                          { max. number of patterns }

type  Layers   = 0..MaxLayer;                         { available layers  }
      Neurons  = 1..MaxNeuron;                        { available neurons }
      NeurThrs = 0..MaxNeuron;      { neurons including thresholds source }
      Patterns = 1..MaxPattern;                       { usable patterns   }
      Weights  = array [Layers,NeurThrs,Neurons] of real;
                       { Weights[i,j,k] :                                 }
                       {   if j>0 ... weight from neuron j in layer i to  }
                       {              neuron k in layer i+1               }
                       {   if j=0 ... threshold of neuron k in layer i+1  }

BP = ^TBP;
TBP = object
      w, wold : Weights;            { values of weights in time t and t-1 }
      x       : array [Layers,NeurThrs] of real;
                        { x[i,j] :                                         }
                        {   if j>0 ... output value of neuron j in layer i }
                        {   if j=0 ... value -1 used as a threshold source }
      Delta   : array [Layers,Neurons] of real;
{ Delta[i,j] = see remark after Eq.(18), concerning now neuron j in layer i }
      NoofL : Layers;                    { layers = 0 [bottom]..NoofL [top] }
      NoofN : array [Layers] of Neurons;  { number of neurons in each layer }
      NoinT : Patterns;                   { number of learning patterns     }
      InPatt, OutPatt  : array [Patterns,Neurons] of real;
                            { all input and expected output patterns from T }
      Input, OutExpect : array [Neurons] of real;
             { input and expected output pattern for one chosen pair from T }
      eta, alpha : real;        { parameters of the algorithm - see Eq.(21) }
      Iters : integer;                               { number of iterations }
      Cycles : integer;                                  { number of cycles }
   constructor Init;
   procedure Run;
   procedure State;                          virtual;
   procedure ChangeParams ( Layer:Layers );  virtual;
   procedure MakeDelta ( Layer:Layers );     virtual;
   procedure InitNetwork;                    virtual;
   procedure InitImpl;                       virtual;
private
   function  S ( Ksi:real ) : real;
   procedure NewWeights;
   function  GlobalError : real;
   procedure Training;
   procedure Testing;
   procedure InitPatterns;
end;


implementation
uses  Crt;               { leave this unit here to speed up screen writings }

constructor TBP.Init;                  { initialization constructor routine }
begin InitNetwork;
      InitImpl;
      InitPatterns;
end;

procedure TBP.Run;               { procedure providing BP algorithm running }
begin Training;
      Testing;
end;

function TBP.S ( Ksi:real ) : real;      { neuron sigmoid transfer function }
const lambda = 1;                                            { sigmoid gain }
      RB     = 30;         { where to extrapolate the sigmoid by a constant }
var   inp : real;
begin inp:=lambda*Ksi;
      if inp>30 then S:=1
                else if inp<-30 then S:=0
                                else S:=1/(1+exp(-inp));
end;

procedure TBP.State;                             { new state of the network }
var   Layer : Layers;
      j     : NeurThrs;
      k     : Neurons;
      Ksi   : real;                                      { neuron potential }
begin for j:=1 to NoofN[0] do
        x[0,j]:=Input[j];                         { set bottom layer inputs }
      for Layer:=1 to NoofL do
        for k:=1 to NoofN[Layer] do
          begin Ksi:=0;
                for j:=0 to NoofN[Layer-1] do
                  Ksi:=Ksi+w[Layer-1,j,k]*x[Layer-1,j];  { neuron potential }
                x[Layer,k]:=S(Ksi)                       { neuron output    }
          end
end;                        { x[NoofL,k] is an actual output of the network }

procedure TBP.ChangeParams ( Layer:Layers );    { new weights for one layer }
var   j     : NeurThrs;
      k     : Neurons;
      saveW : real;
begin for k:=1 to NoofN[Layer+1] do
        for j:=0 to NoofN[Layer] do
          begin saveW:=w[Layer,j,k];
                w[Layer,j,k]:=w[Layer,j,k]-
                              eta*Delta[Layer+1,k]*x[Layer,j] +
                              alpha*(w[Layer,j,k]-Wold[Layer,j,k]);
                Wold[Layer,j,k]:= saveW;
          end;
end;

procedure TBP.MakeDelta ( Layer:Layers );       { new Delta's for one layer }
var   j, k    : Neurons;
      CumulEr : real;            { cumulative error over neurons in a layer }
begin for j:=1 to NoofN[Layer] do
        begin if Layer=NoofL                                    { top layer }
                then CumulEr:=x[NoofL,j]-OutExpect[j]
                else begin CumulEr:=0;      { calculate from previous layer }
                           for k:=1 to NoofN[Layer+1] do
                             CumulEr:=CumulEr+Delta[Layer+1,k]*w[Layer,j,k];
                     end;
              Delta[Layer,j]:=x[Layer,j]*(1-x[Layer,j])*CumulEr
        end
end;

procedure TBP.NewWeights;                             { network new weights }
var   Layer : Layers;
begin for Layer:=NoofL-1 downto 0 do
        begin MakeDelta(Layer+1);  { set up parameters in upper layer }
              ChangeParams(Layer);  { calculate weights in this layer }
        end
end;

function TBP.GlobalError : real;     { global error over all network layers }
var   patt : Patterns;
      j    : Neurons;
      Er   : real;
begin Er:=0;
      for patt:=1 to NoinT do
        begin for j:=1 to NoofN[0] do Input[j]:=InPatt[patt,j];
              for j:=1 to NoofN[NoofL] do OutExpect[j]:=OutPatt[patt,j];
              State;
              for j:=1 to NoofN[NoofL] do
                Er:=Er+Sqr(x[NoofL,j]-OutExpect[j]);
        end;
      GlobalError:=Er;
end;

procedure TBP.Training;                 { provides learning of the patterns }
var   patt  : Patterns;
      j     : Neurons;
      Error : real;                    { cumulative error for one iteration }
      iter, cycle  : integer;
begin
   writeln;                                { format for printed information }
   writeln('Iteration   LayerError   Pattern   Cycle   GlobalError');
   for cycle:=1 to Cycles do
     begin write(chr(13),cycle:38,GlobalError:14:5);     { prints of values }
           for patt:=1 to NoinT do
             begin write(chr(13),patt:29);
                   for j:=1 to NoofN[0] do Input[j]:=InPatt[patt,j];
                   for j:=1 to NoofN[NoofL] do OutExpect[j]:=OutPatt[patt,j];
                   for iter:=1 to Iters do
                     begin State;
                           Error:=0;
                           for j:=1 to NoofN[NoofL] do
                             Error:=Error+Sqr(x[NoofL,j]-OutExpect[j]);
                           NewWeights;
                           write(chr(13),iter:5,Error:16:5);
                     end;
             end;
     end;
   writeln(chr(13),GlobalError:52:5);
end;

procedure TBP.Testing;        { you can try how well the network is learned }
                              { specifying one or more input vectors        }
var   i : Neurons;
      c : char;
begin writeln;
      repeat write('Enter network inputs (',NoofN[0],' values) :  ');
             for i:=1 to NoofN[0] do read(Input[i]);
             readln;
             State;
             write('Output of the network is',':':9);
               for i:=1 to NoofN[NoofL] do write(x[NoofL,i]:5:2);
             write('    More testing [Y/N] ? ');
             read(c);
      until (c='N')or(c='n');
      writeln;
end;

procedure TBP.InitNetwork;   { !! network parameters initialization routine }
var   i : Layers;            { this is the only task dependent procedure !! }
      j : NeurThrs;
      k : Neurons;
      f : text;
begin NoofL:=2;             {the program will deal with the 4-2-4 network }
      NoofN[0]:=4;  NoofN[1]:=2;  NoofN[2]:=4;
      RandSeed:=3456;
{! remove the following brackets numbered 1 if you want to start always  !}
{! with new random weights; if you wish to repeat your experiments       !}
{! always with the same initialization of weights, let them be there     !}
   {1 Randomize; 1}
      for i:=0 to NoofL-1 do
        for j:=0 to NoofN[i] do
          for k:=1 to NoofN[i+1] do
            w[i,j,k]:=6*(Random-0.5)/10;
      Wold:=w;
      eta:=0.1;   alpha:= 0.7;             { choice of learning parameters }
      Iters:= 15; Cycles:=40;  { choice of number of iterations and cycles }

{! remove brackets 2 if you do not want to create your own file of patterns}
{! according to similar template. After removing the brackets 2, you will !}
{! teach the net on identity of vertices of 4-dimensional cube as listed; !}
{! note that the file starts with the number of training pairs.           !}
     { copy patterns into file PATTERNS }
  {2 assign(f,'PATTERNS');
     rewrite(f);
     writeln(f,5);
     writeln(f,'1 1 0 0   1 1 0 0');
     writeln(f,'0 0 1 1   0 0 1 1');
     writeln(f,'1 0 1 0   1 0 1 0');
     writeln(f,'0 1 0 1   0 1 0 1');
     writeln(f,'0 0 0 0   0 0 0 0 ');
     close(f);                        2}
end;

procedure TBP.InitPatterns;                { learning patterns init routine }
var   patt : Patterns;
      j    : Neurons;
      f    : text;
begin assign(f,'PATTERNS'); reset(f);   { using your own file of training set}
      read(f,NoinT); writeln;                           { number of patterns }
      for patt:=1 to NoinT do
        begin for j:=1 to NoofN[0] do
                begin read(f,InPatt[patt,j]);    { read inputs from PATTERNS }
                      write(InPatt[patt,j]:5:2)  { and print them on screen  }
                end;
              write('   ');
              for j:=1 to NoofN[NoofL] do
                begin read(f,OutPatt[patt,j]);  { read outputs from PATTERNS }
                      write(OutPatt[patt,j]:5:2){ and print them on screen   }
                end;
              readln(f); writeln;
        end;
      close(f)
end;

procedure TBP.InitImpl;                       { implementation init routine }
var   Layer : Layers;
begin for Layer:=0 to NoofL-1 do
        x[Layer,0]:=-1;         { used as a threshold source for next layer }
end;

{ begin - no initialization part }
  end.